package org.netkernel.ext.system.accessor;

import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.lang.management.OperatingSystemMXBean;
import java.lang.management.ThreadMXBean;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.container.impl.Kernel;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.Utils;
import org.netkernel.scheduler.RootState;

import com.ten60.netkernel.cache.se.representation2.ConcurrentCache;

public class SystemStatisticsMonitor extends Thread implements IConfigurationListener
{	
	/** housework interval */
	private long mPeriod;
	/** number of memory stats kept */
	private int mWindowSize;
	/** number of housework periods per memory stat period */
	private int mPeriodMultiplier;
	/** true if we have been told to stop */
	private boolean mStop;
	/** maximum memory **/
	private long mMax;
	
	/** the history of memory statistics */
	private long[] mStats;
	/** a smoothed baseline figure to take account of our sampling period error */
	private long mBaseLine;
	/** the last real GC baseline we captured */
	private long mRealBaseLine;
	/** a smoothed peak memory usage to take account of our sampling period error */
	private long mPeak;
	/** the last real GC peak memory we captured */
	private long mRealPeak;
	/** the last memory usage sample we took */
	private long mLast;
	/** total number of requests issued last time checked */
	private long mLastRequests;
	/** counter for implementing frequency divider */
	private int mFreqDivider;
	/** index into rolling statistics buffer **/
	private int mBufferIndex;
	
	private int mCachePeak;
	private int mCacheTrough;
	private float mReqs=10.0f;
	private int mCacheSize;
	private float mHitPercent;
	
	private static final int STAT_COUNT=12;
	
	//private Kernel mKernel;
	
	private ConcurrentCache mCache=null;
	
	public SystemStatisticsMonitor()
	{	super("System Statistics Monitor");
	}
	
	public synchronized void configurationChanged(IConfiguration aConfiguration)
	{
		long period = aConfiguration.getLong("netkernel.statistics.period", 5000L);
		int window = aConfiguration.getInt("netkernel.statistics.window", 60);
		int multiplier = aConfiguration.getInt("netkernel.statistics.multiplier", 10);
		
		if (period!=mPeriod || multiplier!=mPeriodMultiplier)
		{	//restart capture if period has changed
			mStats=null;
		}
		mPeriod=period;
		mPeriodMultiplier=multiplier;
		
		if (mWindowSize!=window || mStats==null)
		{
			if (mStats==null)
			{	mStats =  new long[window*STAT_COUNT];
				mBufferIndex=0;
			}
			else
			{	long[] stats = new long[window*STAT_COUNT];
				int j=mBufferIndex;
				int k=(window-1)*STAT_COUNT;
				for (int i=0; i<mWindowSize && i<window; i++)
				{	System.arraycopy(mStats, j, stats, k, STAT_COUNT);
					k-=STAT_COUNT;
					j-=STAT_COUNT;
					if (j<0) j=(mWindowSize-1)*STAT_COUNT;
				}
				mStats=stats;
				mBufferIndex=(window-1)*STAT_COUNT;
				
			}
			mWindowSize=window;
		}
	}	
	
	public void init(INKFRequestContext context) throws Exception
	{	
		mCache=ConcurrentCache.getInstance();
		
		MemoryPoolMXBean oldGen=Utils.getOldGen();
		MemoryUsage mu=oldGen.getUsage();
		mBaseLine=mu.getUsed();
		mLast=mBaseLine;
		mPeak=mBaseLine;
		mRealPeak=mBaseLine;
		mMax=mu.getMax()>>10;
		this.start();
	}
	
	/** stops its thread */
	public void shutdown()
	{	
		mStop=true;
		this.interrupt();
		mCache=null;
		//mKernel=null;
	}
	
	/** loops until stopped performing work and sleeping between */
	public void run()
	{	long diff=0;
		while (!mStop)
		{	try
			{	long period=mPeriod/mPeriodMultiplier-diff;
				if (period>0)
				{	Thread.sleep(period);
				}
				long t0=System.currentTimeMillis();
				this.doPeriodicHouseKeeping();
				long t1=System.currentTimeMillis();
				diff=t1-t0;
			}
			catch (Throwable e) {}
		}
	}
	
	/** calculate memory statistics by watching for GC events
	 **/
	public synchronized void doPeriodicHouseKeeping()
	{	MemoryPoolMXBean oldGen=Utils.getOldGen();
		MemoryUsage mu=oldGen.getUsage();
		long used = mu.getUsed();
		if (used<mLast)
		{	
			if (used<mRealBaseLine)
			{	mRealBaseLine = used;
			}
			else
			{	mRealBaseLine=used/2 + (mRealBaseLine/2);
			}
			if (mLast>mRealPeak)
			{	mRealPeak = mLast;
			}
			else
			{	mRealPeak=mLast/2 + (mRealPeak/2);
			}
		}
		else
		{	mRealPeak=mRealPeak*31/32+used/32;
		}
		mBaseLine = (mBaseLine*7/8) + mRealBaseLine/8;
		mPeak = (mPeak*7/8) + mRealPeak/8;
		mLast = used;
		
		
		
		mFreqDivider = (mFreqDivider+1)%mPeriodMultiplier;
		if (mFreqDivider==0)
		{	mBufferIndex = (mBufferIndex+STAT_COUNT)%(mWindowSize*STAT_COUNT);
			IHDSNode cacheStatistics=mCache.getStatistics();
			long totalReq=getRequestCount();
			float reqs;
			if (mLastRequests==0 || totalReq<mLastRequests)
			{	reqs=0.0f;
			}
			else
			{	reqs=(totalReq-mLastRequests);
			}
			mLastRequests=totalReq;
			mReqs=reqs; //mReqs*0.9f + reqs*0.1f;
			
			int cachePeakSize=(Integer)cacheStatistics.getFirstValue("/cache/peakSize");
			int cacheTroughSize= (Integer)cacheStatistics.getFirstValue("/cache/troughSize");
			
			if (cachePeakSize>mCachePeak)
			{	mCachePeak=mCachePeak/2 + cachePeakSize/2;
			}
			else
			{	mCachePeak=mCachePeak*31/32+cachePeakSize/32;
			}
			if (cacheTroughSize<mCacheTrough)
			{	mCacheTrough=mCacheTrough/2+cacheTroughSize/2;
			}
			else
			{	mCacheTrough=mCacheTrough*31/32+(cacheTroughSize)/32;
			}
			
			int cacheSize=(Integer)cacheStatistics.getFirstValue("/cache/cacheSize");
			mCacheSize=mCacheSize*15/16+cacheSize/16;
			
			float hitPercent=(Float)cacheStatistics.getFirstValue("/cache/hitPercent");
			mHitPercent=mHitPercent*15/16+hitPercent/16;
			
			long now=System.currentTimeMillis();
			long period=now-mLastNow;
			mLastNow=now;
			int cpus=Runtime.getRuntime().availableProcessors();
			float cpuTime=getCPUTime(cpus);
			
			long gctime=0;
			List<GarbageCollectorMXBean> gcbeans=ManagementFactory.getGarbageCollectorMXBeans();
			for (GarbageCollectorMXBean gcbean: gcbeans)
			{	gctime+=gcbean.getCollectionTime();
			}
			long gctimeInPeriod=gctime-mLastGCTime;
			mLastGCTime=gctime;
			
			float gt=((float)gctimeInPeriod)/((float)(period*cpus));
			gt=mGCTime.next(gt);
			
			float reqsec=1000*mReqs/(float)mPeriod;
			if (reqsec<0.01f) reqsec=0.01f;
			float lr=1.0f+(float)Math.log10(reqsec);
			lr=mLogReq.next(lr);
			//System.out.println(reqsec+" "+lr);
			
			
			mStats[ mBufferIndex+0 ] = now;
			mStats[ mBufferIndex+1 ] = mBaseLine;
			mStats[ mBufferIndex+2 ] = mPeak;
			mStats[ mBufferIndex+3 ] = mu.getCommitted();
			mStats[ mBufferIndex+4 ] = mCachePeak;
			mStats[ mBufferIndex+5 ] = mCacheTrough;
			mStats[ mBufferIndex+6 ] = mCacheSize;
			mStats[ mBufferIndex+7 ] =(long)(mHitPercent*100.0f);
			mStats[ mBufferIndex+8 ] =(long)(mReqs*1000.0f);
			mStats[ mBufferIndex+9 ] = (long)(cpuTime*100);
			mStats[ mBufferIndex+10 ] = (long)(gt*100);
			mStats[ mBufferIndex+11 ] = (long)(lr*1E6);
			
			//additional optimisation, save any pending changes to FastSchematron
			FastSchematron.saveValidationCache();
		}

	}
	
	private float getCPUTime(int aCPUs)
	{	float result=0;
	
		ThreadMXBean threadBean = ManagementFactory.getThreadMXBean( );
		if (threadBean.isThreadCpuTimeSupported())
		{	boolean cpuTimeEnabled=threadBean.isThreadCpuTimeEnabled();
			if (!cpuTimeEnabled) threadBean.setThreadCpuTimeEnabled(true);
			final long DURATION=100;
			long[] ids=threadBean.getAllThreadIds();
			long[] cpuTimes=new long[ids.length];
			long total=0;
			for (int i=0; i<ids.length; i++)
			{	cpuTimes[i]=threadBean.getThreadCpuTime(ids[i]);
			}
			try
			{	Thread.sleep(DURATION);
				for (int i=0; i<ids.length; i++)
				{	long t=threadBean.getThreadCpuTime(ids[i]);
					if (cpuTimes[i]>=0 && t>=0 && cpuTimes[i]<t)
					{	total+=t-cpuTimes[i];
					}
				}
				if (!cpuTimeEnabled) threadBean.setThreadCpuTimeEnabled(false);
				
				double v=(double)total/(double)(aCPUs*DURATION*1000000);
				result=mCPUTime.next((float)v);
			}
			catch (InterruptedException e) {;}
		}
		return result;
	}
	
	
	private long mLastNow;
	private long mLastGCTime;
	
	private SimpleFIR mLogReq=new SimpleFIR(0.0f);
	private SimpleFIR mCPUTime=new SimpleFIR(0.0f);
	private SimpleFIR mGCTime=new SimpleFIR(0.0f);
	private Map<Long,ThreadTime> mThreadTimes=new HashMap<Long,ThreadTime>();
	
	private static class ThreadTime
	{	public long mCPUTime;
	}

	
	/** write memory statistics out as IHDSNode */
	public IHDSNode getBuffer()
	{	HDSBuilder b=new HDSBuilder();
		b.pushNode("memory");
		b.addNode("max",mMax);
		//b.addNode("totalClassesLoaded",mClassLoaderBean.getTotalLoadedClassCount());
		b.addNode("totalDurationMinutes", mPeriod*mWindowSize/60000);
		b.addNode("totalDurationSeconds", mPeriod*mWindowSize/1000);
		int index = mBufferIndex;
		long maxCache=0;
		long lastTime=0;
		for (int i=0; i<mWindowSize; i++)
		{	index+=STAT_COUNT;
			if (index>=mWindowSize*STAT_COUNT) index=0;
			long time=mStats[index+0];
			if (time>lastTime)
			{
				b.pushNode("stat");
					b.addNode("baseline",mStats[index+1]>>10);
					b.addNode("peak",mStats[index+2]>>10);
					b.addNode("alloc",mStats[index+3]>>10);
					Long cacheSize=(Long)mStats[index+4];
					if(cacheSize>maxCache)
					{	maxCache=cacheSize;			
					}
					b.addNode("cacheSize", mStats[index+6]);
					b.addNode("cacheSizePeak", mStats[index+4]);
					b.addNode("cacheSizeTrough", mStats[index+5]);
					b.addNode("cacheHitRate", ((float)mStats[index+7])/100.0f);
					b.addNode("logRequests", mStats[index+11]);
					b.addNode("requests", mStats[index+8]/1000);
					b.addNode("time",time);
					b.addNode("cpu",mStats[index+9]+mStats[index+10]);
					b.addNode("gctime",mStats[index+10]);
				b.popNode();
			}
			lastTime=time;
		}
		b.addNode("maxCache", maxCache);
		b.popNode();
		return b.getRoot();
	}
	
	/** return a filtered estimate of our GC baseline memory usage- i.e how much memory
	 * are we actually holding that cannot be released */
	public final long getBaselineMemory()
	{	return mRealBaseLine;
	}
	
	public final long getMaxMemory()
	{	return mMax<<10;
	}
	
	/** return a filtered estimate of our peak memory usage0-  i.e. how much memory are will
	 * using as a working set
	 */
	public final long getPeakMemory()
	{	return mRealPeak;
	}
	
	public final long getRequestCount()
	{	return RootState.getNextId();
	}
	
	
	private static class SimpleFIR
	{	private static final int SIZE=3;
		private float mData[]=new float[SIZE];
		public SimpleFIR(float aInitialValue)
		{	for (int i=0; i<SIZE; i++)
			{	mData[i]=aInitialValue;
			}
		}
		
		public float next(float aData)
		{	float result=aData*0.4f +mData[0]*0.3f +mData[1]*0.2f + mData[2]*0.1f;
			System.arraycopy(mData, 0, mData, 1, SIZE-1);
			mData[0]=aData;
			return result;
		}
	}
}