package org.netkernel.ext.system.init;

import java.io.File;
import java.io.FileInputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;

import org.netkernel.container.ILogger;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.boot.IModuleManagerSyncListener;
import org.netkernel.layer0.boot.ModuleManager;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.module.standard.endpoint.StandardTransportImpl;

public class InitEndpoint extends StandardTransportImpl implements IConfigurationListener
{	private boolean mBooted;

	
	private static final int LOAD_ALL_LEVELS=-1;
	@Override
	protected void postCommission(INKFRequestContext context) throws Exception
	{	if(!mBooted)
		{	ModuleManager mm=ModuleManager.getSingleton();
	 		mm.registerSyncListener(new InitModuleManagerSyncListener());
	 		context.getKernelContext().getKernel().addConfigurationListener(this);
		}
	}
	
	private void initBoot() throws Exception
	{	mBooted=true;
		getTransportContext().logFormatted(INKFRequestContext.LEVEL_INFO,"MSG_INIT_STARTED", null);
		IConfiguration config=getTransportContext().getKernelContext().getKernel().getConfiguration();
		
		int limitlevel=LOAD_ALL_LEVELS;
		try
		{	String limitlevelstr=config.getString("netkernel.init.runlevel");
			limitlevel=Integer.parseInt(limitlevelstr);
		}
		catch(Exception e){ /*Do nothing */}
		
		if(BootUtils.bootedFromJar(config))
		{	System.out.println("Init not implemented from Jar boot yet")	;
		}
		else
		{	
      ArrayList<File> dirlist = new ArrayList<File>();
      URI base = URI.create(BootUtils.getInstallPath(config));
      URI modules = BootUtils.getKernelURIProperty(config,"netkernel.init.modules");
      if (modules != null) {
        if (!modules.isAbsolute()) {
          modules = base.resolve(modules);
        }
        if (modules.getScheme().equals("file")) {
          File modulesfile = new File(modules);
          if (modulesfile.isFile()) {
            dirlist.add(modulesfile);
          }
        }
      } 
      URI modulesdir = BootUtils.getKernelURIProperty(config,"netkernel.init.modulesdir");
      if(modulesdir==null)
      {		modulesdir=URI.create("etc/modules.d/");		//Default to the generic named directory        	  
      }
      
		if (!modulesdir.isAbsolute()) {
		  modulesdir = base.resolve(modulesdir);  
		}
		if (modulesdir.getScheme().equals("file")) {
		  File modulesdirfile = new File(modulesdir);
		  if (modulesdirfile.isDirectory()) {
		    dirlist.addAll(Arrays.asList(modulesdirfile.listFiles()));
		  }
		}
      

      ModuleManager mm= ModuleManager.getSingleton();
      for (File f : dirlist) {
        if (f.getName().endsWith(".xml")) {
          IHDSNode mods = null;
          IHDSNodeList nl = null;
          int nlsize = 0;
          try {
            mods = HDSFactory.parseXML(new FileInputStream(f));
            nl=mods.getNodes("//module");
            nlsize = nl.size();
          }
          catch (Exception e) {
            getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING,"MSG_MODULESXML_INVALID_FORMAT",f.getAbsolutePath());
          }
          for(int i=0; i<nlsize; i++)
          { IHDSNode n=nl.get(i);
            URI u=base.resolve((String)n.getValue());
            String levels=(String)n.getFirstValue("@runlevel");
            int level=ModuleManager.DEFAULT_RUN_LEVEL;
            if(levels!=null)
            { try
              { level=Integer.parseInt(levels);
              }
              catch(Exception e)
              { getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING, "MSG_RUNLEVEL_INVALID_INTEGER", u.toString());
                level=ModuleManager.DEFAULT_RUN_LEVEL;
              }
              if(level<1)
              { getTransportContext().logFormatted(INKFRequestContext.LEVEL_WARNING,"WARN_INIT_LEVEL_TOO_LOW",new Object[]{u.toString(), ModuleManager.DEFAULT_RUN_LEVEL});
                level=ModuleManager.DEFAULT_RUN_LEVEL;
              }
            }
            mm.addModule(u, level);
          }
        }
      }
      Boolean incremental=config.getBoolean("netkernel.init.incremental", false);
      if(incremental)
      { for(int steplevel=2; steplevel<limitlevel+1; steplevel++)
        { mm.setRunLevel(steplevel);
        }
      }
      else
      { mm.setRunLevel(limitlevel);
      }
      //Start monitoring for changes now we're booted
      mm.startExpiryPollThread();
		}
		getTransportContext().logFormatted(INKFRequestContext.LEVEL_INFO,"MSG_INIT_COMPLETED",new Object[]{limitlevel});
	}
	
	@Override
	protected void preDecommission(INKFRequestContext context) throws Exception
	{
 		context.getKernelContext().getKernel().removeConfigurationListener(this);
	}
	
	class InitModuleManagerSyncListener implements IModuleManagerSyncListener
	{
		public void syncComplete()
		{	ModuleManager.getSingleton().registerSyncListener(null);
			try
			{	initBoot();
			}
			catch(Exception e)
			{	e.printStackTrace();		
			}
		}
	}

	public void configurationChanged(IConfiguration aConfiguration)
	{	
		if (mBooted)
		{
			int runlevel=aConfiguration.getInt("netkernel.init.runlevel", 5);
			ModuleManager.getSingleton().setRunLevel(runlevel);
		}
	}		
}
